# Copyright 2021 Software Factory Labs, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Setup for authorization and secret management for Conduit.

Utilizes Fernet for symmetric encryption of secrets used by Conduit.
"""

import os
import logging
from pathlib import Path
from typing import Any, Generator, Optional, Callable
from contextlib import contextmanager
import yaml
from cryptography.fernet import Fernet

logger = logging.getLogger(__name__)

auth_dir: Optional[Path] = None
secret_path: Optional[Path] = None

secrets: Optional[dict] = None


def _require_setup(func: Callable[..., Any]) -> Any:
    def wrapper(*args: Any, **kwargs: Any) -> Any:
        if auth_dir is None or secret_path is None or secrets is None:
            raise AuthNotSetupException

        return func(*args, **kwargs)

    return wrapper


def setup(etc_dir: Path) -> None:
    """Sets up necessary directories for Conduit, and reads in secrets.

    The `etc/auth/` and `etc/auth/secrets/` directories are created here.

    Args:
        etc_dir (Path): A `Path` to the etc directory.
    """
    logger.debug(f'etc_dir: {etc_dir}')

    # create paths
    global auth_dir
    auth_dir = etc_dir / 'auth'
    auth_dir.mkdir(parents=True, exist_ok=True)
    logger.debug(f'auth_dir: {auth_dir}')

    global secret_path
    secret_path = auth_dir / 'secret'
    logger.debug(f'secret_path: {secret_path}')

    # create key if it does not exist
    if not secret_path.exists():
        logger.debug(f'secret file ({secret_path}) does not exist')
        _create_key_file()
    else:
        with secret_path.open() as f:
            contents = f.read()

        if not contents:
            logger.debug(f'secret file ({secret_path}) empty')
            _create_key_file()

        del contents

        secret_file_perms = os.stat(secret_path).st_mode & 0o777
        desired_perms = 0o600
        if secret_file_perms != desired_perms:
            logging.warning(f'Secret file perms ({oct(secret_file_perms)[2:]}) not secure. '
                            f'Changing to {oct(desired_perms)[2:]}')
            os.chmod(secret_path, desired_perms)

    # load connections config file and encrypt any plaintext passwords
    secrets_file_path = etc_dir / 'local' / 'secrets.yaml'

    logger.debug('loading secrets')
    global secrets
    
    if secrets_file_path.exists():
        with secrets_file_path.open() as file:
            secrets = yaml.safe_load(file)

    if secrets is None:
        secrets = {}

    change = False

    for secret_id, secret_text in secrets.items():
        if secret_text.startswith('$1$'):
            continue

        logger.info(f'Encrypting secret for "{secret_id}"')

        secrets[secret_id] = f'$1${_encrypt(secret_text).decode()}'
        del secret_text

        change = True

    if change:
        with secrets_file_path.open('w') as f:
            yaml.safe_dump(secrets, f)

    logger.debug(f'loaded secrets: {list(secrets.keys())}')


@_require_setup
@contextmanager
def get_secret(secret_id: str) -> Generator[str, None, None]:
    """Retrieves a secret temporarily.

    Works as a context manager, where the secret will be deleted from memory upon
    exiting.

    Args:
        secret_id (str): The ID of the secret to get.

    Raises:
        Exception: When no secret was associated with the `secret_id`.

    Yields:
        Iterator[str]: The decrypted secret.
    """
    try:
        # just for mypy; should never happen because of the @_require_setup decorator
        if secrets is None:
            raise AuthNotSetupException

        encrypted_password = secrets.get(secret_id, None)

        if encrypted_password is None:
            raise NoSecretFoundException(secret_id)

        encrypted_password = encrypted_password.split('$')[2]

        password = _decrypt(encrypted_password.encode())
        try:
            yield password
        finally:
            del password
    except Exception as e:
        logger.exception(e)
        raise


def _create_key_file() -> None:
    """Creates a secret key file."""
    # just for mypy; should never happen because this function gets called during setup()
    if secret_path is None:
        raise AuthNotSetupException

    logger.info('Creating secret key')
    with secret_path.open('wb') as f:
        key = Fernet.generate_key()
        f.write(key)
        del key

    os.chmod(secret_path, 0o400)


@_require_setup
def _encrypt(plaintext: str) -> bytes:
    """Encrypts the given text.

    Args:
        plaintext (str): Text to encrypt.

    Returns:
        bytes: The encrypted text.
    """
    with _get_fernet() as f:
        return f.encrypt(plaintext.encode())


@_require_setup
def _decrypt(ciphertext: bytes) -> str:
    """Decrypt the given ciphertext.

    Args:
        ciphertext (bytes): Encrypted text to decrypt.

    Returns:
        str: The decrypted text.
    """
    with _get_fernet() as f:
        return f.decrypt(ciphertext).decode()


@_require_setup
@contextmanager
def _get_fernet() -> Generator[Fernet, None, None]:
    """Creates a `Fernet` to allow for encryption and decryption using the secret key.

    Returns:
        Fernet: Allows for encryption and decryption utilizing your secret key.

    Yields:
        Iterator[Fernet]: The initialized `Fernet`, which is deleted from memory after use.
    """
    # just for mypy; should never happen because this function gets called during setup()
    if secret_path is None:
        raise AuthNotSetupException

    with secret_path.open() as f:
        key = f.read()

    fern = Fernet(key)  # type: ignore

    try:
        yield fern
    finally:
        del fern
        del key


class AuthNotSetupException(Exception):
    """Raised when authorization and secrets not setup yet."""
    pass


class NoSecretFoundException(Exception):
    """Raised when the requested secret is not found."""
    def __init__(self, secret_id: str):
        super().__init__(f'No secret found for "{secret_id}"')
